# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import json
from dataclasses import dataclass, asdict
from typing import Optional

from eth_account.messages import encode_defunct
from pydantic import BaseModel
from web3 import Web3

import vana
from vana.utils import networking
from vana.utils.web3 import as_wad


@dataclass
class NodeServerInfo:
    version: int
    ip: str
    port: int
    ip_type: int
    hotkey: str
    coldkey: str

    @property
    def is_serving(self) -> bool:
        """True if the endpoint is serving."""
        if self.ip == "0.0.0.0":
            return False
        else:
            return True

    def ip_str(self) -> str:
        """Return the whole IP as string"""
        return networking.ip__str__(self.ip_type, self.ip, self.port)

    def __eq__(self, other: "NodeServerInfo"):
        if other == None:
            return False
        if (
                self.version == other.version
                and self.ip == other.ip
                and self.port == other.port
                and self.ip_type == other.ip_type
                and self.coldkey == other.coldkey
                and self.hotkey == other.hotkey
        ):
            return True
        else:
            return False

    def __str__(self):
        return "NodeServerInfo( {}, {}, {}, {} )".format(
            str(self.ip_str()), str(self.hotkey), str(self.coldkey), self.version
        )

    def __hash__(self):
        return hash((self.version, self.ip, self.port, self.ip_type, self.hotkey, self.coldkey))

    def __repr__(self):
        return self.__str__()

    def to_string(self) -> str:
        """Converts the NodeServerInfo object to a string representation using JSON."""
        try:
            return json.dumps(asdict(self))
        except (TypeError, ValueError) as e:
            vana.logging.error(f"Error converting NodeServerInfo to string: {e}")
            return NodeServerInfo(0, "", 0, 0, "", "").to_string()

    @classmethod
    def from_string(cls, s: str) -> "NodeServerInfo":
        """Creates an NodeServerInfo object from its string representation using JSON."""
        try:
            data = json.loads(s)
            return cls(**data)
        except json.JSONDecodeError as e:
            vana.logging.error(f"Error decoding JSON: {e}")
        except TypeError as e:
            vana.logging.error(f"Type error: {e}")
        except ValueError as e:
            vana.logging.error(f"Value error: {e}")
        return NodeServerInfo(0, "", 0, 0, "", "")


class ProofData(BaseModel):
    file_url: str
    score: float
    dlp_id: int
    metadata: Optional[str] = ""
    proof_url: Optional[str] = ""
    instruction: str


class Proof(BaseModel):
    signature: Optional[str] = ""
    data: ProofData

    def sign(self, wallet):
        packed_data = Web3().solidity_keccak(
            ['string', 'uint256', 'uint256', 'string', 'string', 'string'],
            [
                self.data.file_url,
                as_wad(self.data.score),
                self.data.dlp_id,
                self.data.metadata,
                self.data.proof_url,
                self.data.instruction
            ]
        )

        message = encode_defunct(packed_data)
        self.signature = wallet.hotkey.sign_message(message).signature.hex()
        return self



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import logging


def all_loggers():
    """
    Generator that yields all logger instances in the application.
    """
    for logger in logging.root.manager.loggerDict.values():
        if isinstance(logger, logging.PlaceHolder):
            continue
        # In some versions of Python, the values in loggerDict might be
        # LoggerAdapter instances instead of Logger instances.
        # We check for Logger instances specifically.
        if isinstance(logger, logging.Logger):
            yield logger
        else:
            # If it's not a Logger instance, it could be a LoggerAdapter or
            # another form that doesn't directly offer logging methods.
            # This branch can be extended to handle such cases as needed.
            pass


def all_logger_names():
    for name, logger in logging.root.manager.loggerDict.items():
        if isinstance(logger, logging.PlaceHolder):
            continue
        # In some versions of Python, the values in loggerDict might be
        # LoggerAdapter instances instead of Logger instances.
        # We check for Logger instances specifically.
        if isinstance(logger, logging.Logger):
            yield name
        else:
            # If it's not a Logger instance, it could be a LoggerAdapter or
            # another form that doesn't directly offer logging methods.
            # This branch can be extended to handle such cases as needed.
            pass


def get_max_logger_name_length():
    max_length = 0
    for name in all_logger_names():
        if len(name) > max_length:
            max_length = len(name)
    return max_length



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import asyncio
import time
import uuid
from typing import Optional, Union, List, AsyncGenerator, Any

import aiohttp
from eth_account.messages import encode_defunct

import vana
from vana.utils.networking import get_external_ip


class NodeClient:
    """
    A client for interacting with a node's server.
    """

    def __init__(
            self, wallet: Optional[vana.Wallet] = None
    ):
        self.uuid = str(uuid.uuid1())
        self.external_ip = get_external_ip()
        self.keypair = (wallet.hotkey if isinstance(wallet, vana.Wallet) else wallet) or vana.Wallet().hotkey
        self.message_history: list = []
        self._session: Optional[aiohttp.ClientSession] = None

    @property
    async def session(self) -> aiohttp.ClientSession:
        """
        An asynchronous property that provides access to the internal `aiohttp <https://github.com/aio-libs/aiohttp>`_ client session.
        Example usage::
            node_client = NodeClient()
            async with (await node_client.session).post( # Use the session to make an HTTP POST request
                url,                                  # URL to send the request to
                headers={...},                        # Headers dict to be sent with the request
                json={...},                           # JSON body data to be sent with the request
                timeout=10,                           # Timeout duration in seconds
            ) as response:
                json_response = await response.json() # Extract the JSON response from the server
        """
        if self._session is None:
            self._session = aiohttp.ClientSession()
        return self._session

    def close_session(self):
        """
        Closes the internal `aiohttp <https://github.com/aio-libs/aiohttp>`_ client session synchronously.
        """
        if self._session:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self._session.close())
            self._session = None

    async def aclose_session(self):
        """
        Asynchronously closes the internal `aiohttp <https://github.com/aio-libs/aiohttp>`_ client session.
        """
        if self._session:
            await self._session.close()
            self._session = None

    def _get_endpoint_url(self, target_node_server, request_name):
        """
        Constructs the endpoint URL for a network request to a target NodeServer.
        """
        endpoint = (
            f"0.0.0.0:{str(target_node_server.port)}"
            if target_node_server.ip == str(self.external_ip)
            else f"{target_node_server.ip}:{str(target_node_server.port)}"
        )
        return f"http://{endpoint}/{request_name}"

    def _handle_request_errors(self, message, request_name, exception):
        """
        Handles exceptions that occur during network requests, updating the message with appropriate status codes and messages.

        This method interprets different types of exceptions and sets the corresponding status code and
        message in the message object. It covers common network errors such as connection issues and timeouts.

        Args:
            message: The message object associated with the request.
            request_name: The name of the request during which the exception occurred.
            exception: The exception object caught during the request.

        Note:
            This method updates the message object in-place.
        """
        if isinstance(exception, aiohttp.ClientConnectorError):
            message.node_client.status_code = "503"
            message.node_client.status_message = f"Service at {message.node_server.ip}:{str(message.node_server.port)}/{request_name} unavailable."
        elif isinstance(exception, asyncio.TimeoutError):
            message.node_client.status_code = "408"
            message.node_client.status_message = (
                f"Timedout after {message.timeout} seconds."
            )
        else:
            message.node_client.status_code = "422"
            message.node_client.status_message = (
                f"Failed to parse response object with error: {str(exception)}"
            )

    def _log_outgoing_request(self, message):
        """
        Logs information about outgoing requests for debugging purposes.
        """
        vana.logging.trace(
            f"node_client | --> | {message.get_total_size()} B | {message.name} | {message.node_server.hotkey} | {message.node_server.ip}:{str(message.node_server.port)} | 0 | Success"
        )

    def _log_incoming_response(self, message):
        """
        Logs information about incoming responses for debugging and monitoring.
        """
        vana.logging.trace(
            f"node_client | <-- | {message.get_total_size()} B | {message.name} | {message.node_server.hotkey} | {message.node_server.ip}:{str(message.node_server.port)} | {message.node_client.status_code} | {message.node_client.status_message}"
        )

    def query(self, *args, **kwargs) -> List[Union[AsyncGenerator[Any, Any], vana.Message]]:
        """
        Makes a synchronous request to multiple target NodeServers and returns the responses.

        Cleanup is automatically handled and sessions are closed upon completed requests.

        Args:
            node_servers (Union[List[Union['NodeServerInfo', 'NodeServer']], Union['NodeServerInfo', 'NodeServer']]):
                The list of target NodeServer information.
            message (Message, optional): The Message object. Defaults to :func:`Message()`.
            timeout (float, optional): The request timeout duration in seconds.
                Defaults to ``12.0`` seconds.
        Returns:
            Union[Message, List[Message]]: If a single target node_server is provided, returns the response from that node_server. If multiple target node_servers are provided, returns a list of responses from all target node_servers.
        """
        result = None
        try:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(self.forward(*args, **kwargs))
        except:
            new_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(new_loop)
            result = new_loop.run_until_complete(self.forward(*args, **kwargs))
            new_loop.close()
        finally:
            self.close_session()
            return result

    async def forward(
            self,
            node_servers: Union[
                List[Union[vana.NodeServerInfo, vana.NodeServer]],
                Union[vana.NodeServerInfo, vana.NodeServer],
            ],
            message: vana.Message = vana.Message(),
            timeout: float = 60,
            deserialize: bool = True,
            run_async: bool = True,
    ) -> List[
        Union[AsyncGenerator[Any, Any], vana.Message]
    ]:
        """
        Asynchronously sends requests to one or multiple NodeServers and collates their responses.
        """
        is_list = True
        # If a single node_servers is provided, wrap it in a list for uniform processing
        if not isinstance(node_servers, list):
            is_list = False
            node_servers = [node_servers]

        async def query_all_node_servers() -> Union[AsyncGenerator[Any, Any], vana.Message]:
            """
            Handles the processing of requests to all targeted node_servers
            """

            async def single_node_server_response(target_node_server) -> Union[AsyncGenerator[Any, Any], vana.Message]:
                """
                Manages the request and response process for a single node_server
                """
                return await self.call(
                    target_node_server=target_node_server,
                    message=message.copy(),
                    timeout=timeout,
                    deserialize=deserialize,
                )

            # If run_async flag is False, get responses one by one.
            if not run_async:
                return [await single_node_server_response(target_node_server) for target_node_server in node_servers]
            # If run_async flag is True, get responses concurrently using asyncio.gather().
            return await asyncio.gather(
                *(single_node_server_response(target_node_server) for target_node_server in node_servers))

        # Get responses for all node_servers.
        responses = await query_all_node_servers()
        # Return the single response if only one node_server was targeted, else return all responses
        return responses[0] if len(responses) == 1 and not is_list else responses

    async def call(
            self,
            target_node_server: Union[vana.NodeServerInfo, vana.NodeServer],
            message: vana.Message = vana.Message(),
            timeout: float = 12.0,
            deserialize: bool = True,
    ) -> vana.Message:
        """
        Asynchronously sends a request to a specified NodeServer and processes the response.
        """

        # Record start time
        start_time = time.time()
        target_node_server = (
            target_node_server.info()
            if isinstance(target_node_server, vana.NodeServer)
            else target_node_server
        )

        # Build request endpoint from the message class
        request_name = message.__class__.__name__
        url = self._get_endpoint_url(target_node_server, request_name=request_name)

        # Preprocess message for making a request
        message = self.preprocess_message_for_request(target_node_server, message, timeout)

        try:
            # Log outgoing request
            self._log_outgoing_request(message)

            # Make the HTTP POST request
            async with (await self.session).post(
                    url,
                    headers=message.to_headers(),
                    json=message.dict(),
                    timeout=timeout,
            ) as response:
                # Extract the JSON response from the server
                json_response = await response.json()
                # Process the server response and fill message
                self.process_server_response(response, json_response, message)

            # Set process time and log the response
            message.node_client.process_time = str(time.time() - start_time)

        except Exception as e:
            self._handle_request_errors(message, request_name, e)

        finally:
            self._log_incoming_response(message)

            # Log message event history
            self.message_history.append(
                vana.Message.from_headers(message.to_headers())
            )

            # Return the updated message object after deserializing if requested
            if deserialize:
                return message.deserialize()
            else:
                return message

    def preprocess_message_for_request(
            self,
            target_node_server_info: vana.NodeServerInfo,
            message: vana.Message,
            timeout: float = 12.0,
    ) -> vana.Message:
        """
        Preprocesses the message for making a request. This includes building
        headers for NodeClient and NodeServer and signing the request.

        Args:
            target_node_server_info (NodeServerInfo): The target NodeServer information.
            message (Message): The message object to be preprocessed.
            timeout (float, optional): The request timeout duration in seconds.
                Defaults to ``12.0`` seconds.

        Returns:
            Message: The preprocessed message.
        """
        # Set the timeout for the message
        message.timeout = timeout

        # Build the NodeClient headers using the local system's details
        message.node_client = vana.TerminalInfo(
            ip=self.external_ip,
            version=vana.__version_as_int__,
            nonce=time.monotonic_ns(),
            uuid=self.uuid,
            hotkey=self.keypair.address,
        )

        # Build the NodeServer headers using the target NodeServer's details
        message.node_server = vana.TerminalInfo(
            ip=target_node_server_info.ip,
            port=target_node_server_info.port,
            hotkey=target_node_server_info.hotkey,
        )

        # Sign the request using the node_client, NodeServer info, and the message body hash
        message_to_sign = f"{message.node_client.nonce}.{message.node_client.hotkey}.{message.node_server.hotkey}.{message.node_client.uuid}.{message.body_hash}"
        signable_message = encode_defunct(text=message_to_sign)
        message.node_client.signature = f"0x{self.keypair.sign_message(signable_message).signature}"

        return message

    def process_server_response(
            self,
            server_response: aiohttp.ClientResponse,
            json_response: dict,
            local_message: vana.Message,
    ):
        """
        Processes the server response, updates the local message state with the
        server's state and merges headers set by the server.

        Args:
            server_response (object): The `aiohttp <https://github.com/aio-libs/aiohttp>`_ response object from the server.
            json_response (dict): The parsed JSON response from the server.
            local_message (Message): The local message object to be updated.

        Raises:
            None: But errors in attribute setting are silently ignored.
        """
        # Check if the server responded with a successful status code
        if server_response.status == 200:
            # If the response is successful, overwrite local message state with
            # server's state only if the protocol allows mutation. To prevent overwrites,
            # the protocol must set allow_mutation = False
            server_message = local_message.__class__(**json_response)
            for key in local_message.dict().keys():
                try:
                    # Set the attribute in the local message from the corresponding
                    # attribute in the server message
                    setattr(local_message, key, getattr(server_message, key))
                except:
                    # Ignore errors during attribute setting
                    pass

        # Extract server headers and overwrite None values in local message headers
        server_headers = vana.Message.from_headers(server_response.headers)

        # Merge node_client headers
        local_message.node_client.__dict__.update(
            {
                **local_message.node_client.dict(exclude_none=True),
                **server_headers.node_client.dict(exclude_none=True),
            }
        )

        # Merge NodeServer headers
        local_message.node_server.__dict__.update(
            {
                **local_message.node_server.dict(exclude_none=True),
                **server_headers.node_server.dict(exclude_none=True),
            }
        )

        # Update the status code and status message of the node_client to match the NodeServer
        local_message.node_client.status_code = server_response.status
        local_message.node_client.status_message = local_message.node_server.status_message

    def __str__(self) -> str:
        """
        Returns a string representation of the NodeClient object.

        Returns:
            str: The string representation of the NodeClient object in the format :func:`node_client(<user_wallet_address>)`.
        """
        return "node_client({})".format(self.keypair.address)

    def __repr__(self) -> str:
        """
        Returns a string representation of the NodeClient object, acting as a fallback for :func:`__str__()`.

        Returns:
            str: The string representation of the NodeClient object in the format :func:`node_client(<user_wallet_address>)`.
        """
        return self.__str__()

    async def __aenter__(self):
        """
        Asynchronous context manager entry method.

        Enables the use of the ``async with`` statement with the NodeClient instance. When entering the context,
        the current instance of the class is returned, making it accessible within the asynchronous context.

        Returns:
            NodeClient: The current instance of the NodeClient class.

        Usage::

            async with NodeClient() as node_client:
                await node_client.some_async_method()
        """
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """
        Asynchronous context manager exit method.

        Ensures proper cleanup when exiting the ``async with`` context. This method will close the `aiohttp <https://github.com/aio-libs/aiohttp>`_ client session
        asynchronously, releasing any tied resources.

        Args:
            exc_type (Type[BaseException], optional): The type of exception that was raised.
            exc_value (BaseException, optional): The instance of exception that was raised.
            traceback (TracebackType, optional): A traceback object encapsulating the call stack at the point where the exception was raised.

        Usage::

            async with node_client( wallet ) as node_client:
                await node_client.some_async_method()

        Note:
            This automatically closes the session by calling :func:`__aexit__` after the context closes.
        """
        await self.aclose_session()

    def __del__(self):
        """
        NodeClient destructor.

        This method is invoked when the NodeClient instance is about to be destroyed. The destructor ensures that the
        aiohttp client session is closed before the instance is fully destroyed, releasing any remaining resources.

        Note:
            Relying on the destructor for cleanup can be unpredictable. It is recommended to explicitly close sessions using the provided methods or the ``async with`` context manager.

        Usage::

            node_client = NodeClient()
            # ... some operations ...
            del node_client  # This will implicitly invoke the __del__ method and close the session.
        """
        self.close_session()



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import copy
import json
import os
import uuid
from inspect import signature, Signature
from typing import Callable, Optional, Dict, List

import uvicorn
from fastapi import FastAPI, APIRouter, Depends, Request

import vana
from vana.utils.fast_api_threaded_server import FastAPIThreadedServer
from vana.utils.networking import get_external_ip


class NodeServer:
    """
     The ``NodeServer`` class serves as the server-side interface for a node within the network.

     Relies on a FastAPI router to create endpoints for different message types.

     Example Usage::
        # Define your custom message class
        class MyMessage( Message ):
            input: int = 1
            output: int = None

        # Define a custom request forwarding function using your message class
        def forward( message: MyMessage ) -> MyMessage:
            message.output = 2
            return message

        # Initialize NodeServer object with a custom configuration
        node_server = NodeServer(
            config=my_config,
            wallet=my_wallet,
            port=9090,
            ip="192.0.2.0",
            external_ip="203.0.113.0",
            external_port=7070
        )

        # Attach the endpoint with the specified verification and forward functions.
        node_server.attach(
            forward_fn = forward_my_message
        )

        # Serve and start your NodeServer.
        node_server.serve().start()
    """

    def __init__(self,
                 wallet: Optional["vana.Wallet"] = None,
                 config: Optional["vana.Config"] = None,
                 port: Optional[int] = None,
                 ip: Optional[str] = None,
                 external_ip: Optional[str] = None,
                 external_port: Optional[int] = None,
                 max_workers: Optional[int] = None):
        if config is None:
            config = self.config()
        config = copy.deepcopy(config)
        config.node_server.ip = ip or config.node_server.get("ip", "0.0.0.0")
        config.node_server.port = port or config.node_server.get("port", 4000)
        config.node_server.external_ip = external_ip or config.node_server.get("external_ip", get_external_ip())
        config.node_server.external_port = external_port or config.node_server.get("external_port", 4000)
        config.node_server.max_workers = max_workers or config.node_server.get("max_workers", 5)
        self.check_config(config)
        self.config = config
        self.wallet = wallet

        # Build NodeServer objects.
        self.uuid = str(uuid.uuid1())
        self.ip = self.config.node_server.ip
        self.port = self.config.node_server.port
        self.external_ip = (
            self.config.node_server.external_ip
            if self.config.node_server.external_ip is not None
            else get_external_ip()
        )
        self.external_port = (
            self.config.node_server.external_port
            if self.config.node_server.external_port is not None
            else self.config.node_server.port
        )
        self.full_address = str(self.ip) + ":" + str(self.port)
        self.full_external_address = str(self.external_ip) + ":" + str(self.external_port)
        self.started = False

        # Request default functions.
        self.forward_class_types: Dict[str, List[Signature]] = {}

        # Instantiate FastAPI
        self.app = FastAPI()
        log_level = "info"
        self.fast_config = uvicorn.Config(
            self.app, host="0.0.0.0", port=self.config.node_server.port, log_level=log_level, loop='asyncio',
            workers=config.node_server.max_workers
        )
        self.fast_server = FastAPIThreadedServer(config=self.fast_config)  # uvicorn.Server(self.fast_config)
        self.router = APIRouter()
        self.app.include_router(self.router)

        # Attach default forward.
        def ping(r: vana.Message) -> vana.Message:
            return r

        self.attach(
            forward_fn=ping
        )

        self.attach_cli_input()

    def info(self) -> "vana.NodeServerInfo":
        """Returns the NodeServerInfo object associated with this NodeServer."""
        return vana.NodeServerInfo(
            version=vana.__version_as_int__,
            ip=self.external_ip,
            ip_type=4,
            port=self.external_port,
            hotkey=self.wallet.hotkey.address,
            coldkey=self.wallet.coldkeypub_str,
        )

    def attach(self, forward_fn: Callable) -> "NodeServer":
        """
        Attaches custom functions to the NodeServer for handling incoming requests.
        Registers an API endpoint to the FastAPI application router.
        """
        # Assert 'forward_fn' has exactly one argument
        forward_sig = signature(forward_fn)
        assert (
                len(list(forward_sig.parameters)) == 1
        ), "The passed function must have exactly one argument"

        # Obtain the class of the first argument of 'forward_fn'
        request_class = forward_sig.parameters[
            list(forward_sig.parameters)[0]
        ].annotation

        # Assert that the first argument of 'forward_fn' is a subclass of 'Message'
        assert issubclass(
            request_class, vana.Message
        ), "The argument of forward_fn must inherit from Message"

        # Obtain the class name of the first argument of 'forward_fn'
        request_name = forward_sig.parameters[
            list(forward_sig.parameters)[0]
        ].annotation.__name__

        # Add the endpoint to the router, making it available on both GET and POST methods
        dependencies = [Depends(self.verify_body_integrity)] if self.config.node_server.verify_body_integrity else []
        self.router.add_api_route(
            f"/{request_name}",
            forward_fn,
            methods=["GET", "POST"],
            dependencies=dependencies,
        )
        self.app.include_router(self.router)

        self.forward_class_types[request_name] = forward_sig.parameters[
            list(forward_sig.parameters)[0]
        ].annotation

        return self

    def attach_cli_input(self):
        """
        Temporary route to send commands to the NodeServer via API.
        Creates a file which the validator will pick up, process, and delete.
        """

        def handle_cli_input(r: Request):
            cli_input = r.query_params.get('input')
            if cli_input is not None:
                try:
                    json.loads(cli_input)
                    with open('cli.json', 'w') as f:
                        f.write(cli_input)
                        return {"status": "success"}
                except json.JSONDecodeError:
                    return {"status": "failed", "message": "invalid json"}
            else:
                return {"status": "failed", "message": "missing input query param"}

        self.router.add_api_route(
            f"/cli",
            handle_cli_input,
            methods=["GET"]
        )
        self.app.include_router(self.router)

    @classmethod
    def config(cls) -> vana.Config:
        """
        Parses the command-line arguments to form a configuration object.
        """
        parser = argparse.ArgumentParser()
        NodeServer.add_args(parser)
        return vana.Config(parser, args=[])

    @classmethod
    def add_args(cls, parser: argparse.ArgumentParser, prefix: Optional[str] = None):
        """
        Adds NodeServer-specific command-line arguments to the argument parser.
        """
        prefix_str = "" if prefix is None else prefix + "."
        try:
            # Get default values from environment variables or use default values
            default_node_server_port = os.getenv("NODESERVER_PORT") or 8091
            default_node_server_ip = os.getenv("NODESERVER_IP") or "[::]"
            default_node_server_external_port = os.getenv("NODESERVER_EXTERNAL_PORT") or None
            default_node_server_external_ip = os.getenv("NODESERVER_EXTERNAL_IP") or None
            default_node_server_max_workers = os.getenv("NODESERVER_MAX_WORKERS") or 10

            # Add command-line arguments to the parser
            parser.add_argument(
                "--" + prefix_str + "node_server.port",
                type=int,
                help="The local port this NodeServer endpoint is bound to. i.e. 8091",
                default=default_node_server_port,
            )
            parser.add_argument(
                "--" + prefix_str + "node_server.ip",
                type=str,
                help="""The local ip this NodeServer binds to. ie. [::]""",
                default=default_node_server_ip,
            )
            parser.add_argument(
                "--" + prefix_str + "node_server.external_port",
                type=int,
                required=False,
                help="""The public port this NodeServer broadcasts to the network. i.e. 8091""",
                default=default_node_server_external_port,
            )
            parser.add_argument(
                "--" + prefix_str + "node_server.external_ip",
                type=str,
                required=False,
                help="""The external ip this NodeServer broadcasts to the network to. ie. [::]""",
                default=default_node_server_external_ip,
            )
            parser.add_argument(
                "--" + prefix_str + "node_server.max_workers",
                type=int,
                help="""The maximum number connection handler threads working simultaneously on this endpoint.
                            The grpc server distributes new worker threads to service requests up to this number.""",
                default=default_node_server_max_workers,
            )
            parser.add_argument(
                "--" + prefix_str + "node_server.verify_body_integrity",
                type=bool,
                help="""Whether to verify the integrity of the body of incoming HTTP requests.""",
                default=True,
            )

        except argparse.ArgumentError:
            # Exception handling for re-parsing arguments
            pass

    async def verify_body_integrity(self, request: Request):
        """
        Responsible for ensuring the integrity of the body of incoming HTTP requests.
        """
        # Await and load the request body, so we can inspect it
        body = await request.body()
        request_body = body.decode() if isinstance(body, bytes) else body
        request_name = request.url.path.split("/")[1]

        # Load the body dict and check if all required field hashes match
        body_dict = json.loads(request_body)

        # Reconstruct the message object from the body dict and recompute the hash
        syn = self.forward_class_types[request_name](**body_dict)  # type: ignore
        parsed_body_hash = syn.body_hash  # Rehash the body from request

        body_hash = request.headers.get("computed_body_hash", "")
        if parsed_body_hash != body_hash:
            raise ValueError(
                f"Hash mismatch between header body hash {body_hash} and parsed body hash {parsed_body_hash}"
            )

        # If body is good, return the parsed body so that it can be passed onto the route function
        return body_dict

    @classmethod
    def check_config(cls, config: "vana.config"):
        """
        This method checks the configuration for the NodeServer's port and wallet.
        """
        pass

    def start(self) -> "NodeServer":
        self.fast_server.start()  # If using not FastAPIThreadedServer, use self.fast_server.run()
        self.started = True
        return self

    def stop(self) -> "NodeServer":
        # self.fast_server.should_exit = True  # If using FastAPIThreadedServer, use self.fast_server.stop()
        self.fast_server.stop()
        self.started = False
        return self

    def serve(self, dlp_uid: Optional[int] = None, chain_manager: Optional["vana.ChainManager"] = None) -> "NodeServer":
        """
        Registers the NodeServer with a specific DLP within the Vana network, identified by the ``dlp_uid``.
        """
        if chain_manager is not None and hasattr(chain_manager, "serve_node_server"):
            chain_manager.serve_node_server(dlp_uid=dlp_uid, node_server=self)
        return self

    def unserve(self, dlp_uid: int, chain_manager: Optional["vana.ChainManager"] = None) -> "NodeServer":
        """
        De-registers the NodeServer with a specific DLP within the Vana network, identified by the ``dlp_uid``.
        """
        if chain_manager is not None and hasattr(chain_manager, "remove_node_server"):
            chain_manager.remove_node_server(dlp_uid=dlp_uid, node_server=self)
        return self



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Union

from eth_abi import decode
from web3 import Web3
from web3.types import ABI


def decode_custom_error(contract_abi: ABI, error_data: Union[str, bytes]) -> str:
    """
    Decodes a custom contract error using the contract ABI.

    Parameters:
    contract_abi (ABI): The ABI of the contract containing error definitions.
    error_data (Union[str, bytes]): The error data returned from a contract call. This can be a hex string or bytes.

    Returns:
    str: A human-readable string representing the decoded error message, or "Unknown error" if the error cannot be decoded.
    """
    if isinstance(error_data, str):
        error_data = Web3.to_bytes(hexstr=error_data)

    # Extract the error signature (first 4 bytes of the error data)
    error_sig = error_data[:4].hex()

    for item in contract_abi:
        if item['type'] == 'error':
            # Construct the full error signature from the ABI definition
            inputs = ','.join(input['type'] for input in item['inputs'])
            full_signature = f"{item['name']}({inputs})"
            calculated_sig = Web3.keccak(text=full_signature)[:4].hex()

            # Compare and match the calculated signature with the error signature from the data
            if calculated_sig[2:] == error_sig:
                error_abi = item
                error_args = error_data[4:]
                error_decoded = decode([input['type'] for input in error_abi['inputs']], error_args)
                error_message = f"{error_abi['name']}({', '.join(map(str, error_decoded))})"
                return error_message

    return f"Unknown error({error_data})"


def as_wad(num: float = 0) -> int:
    """
    Convert a number to its equivalent in wei.
    :param num:
    :return:
    """
    return int(num * 1e18)


def from_wad(num: int = 0) -> float:
    """
    Convert a number from wei to its equivalent.
    :param num:
    :return:
    """
    return num / 1e18



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Union, Optional

from eth_keys import keys
from eth_utils import decode_hex
from web3 import Web3


def is_valid_secp256k1_pubkey(public_key: Union[str, bytes]) -> bool:
    """
    Checks if the given public_key is a valid Ethereum secp256k1 key.

    Args:
        public_key (Union[str, bytes]): The public_key to check.

    Returns:
        bool: True if the public_key is a valid secp256k1 key, False otherwise.
    """
    try:
        if isinstance(public_key, str):
            if public_key.startswith("0x04") and len(public_key) == 130:
                public_key = public_key[2:]
            elif len(public_key) != 128:
                return False
            public_key_bytes = decode_hex(public_key)
        elif isinstance(public_key, bytes):
            if len(public_key) != 64:
                return False
            public_key_bytes = public_key
        else:
            return False

        # Attempt to create a public key object
        keys.PublicKey(public_key_bytes)
        return True

    except (ValueError, TypeError):
        return False


def is_valid_vana_address_or_public_key(address: Union[str, bytes]) -> bool:
    """
    Checks if the given address is a valid Ethereum address or public key.

    Args:
        address (Union[str, bytes]): The address to check.

    Returns:
        bool: True if the address is a valid Ethereum address or public key, False otherwise.
    """
    if isinstance(address, str):
        return Web3.is_address(address) or is_valid_secp256k1_pubkey(address)
    elif isinstance(address, bytes):
        return is_valid_secp256k1_pubkey(address)
    return False


def create_identity_dict(
        display: str = "",
        legal: str = "",
        web: str = "",
        riot: str = "",
        email: str = "",
        pgp_fingerprint: Optional[str] = None,
        image: str = "",
        info: str = "",
        twitter: str = "",
) -> dict:
    """
    Creates a dictionary with structure for identity extrinsic. Must fit within 64 bits.

    Args:
    display (str): String to be converted and stored under 'display'.
    legal (str): String to be converted and stored under 'legal'.
    web (str): String to be converted and stored under 'web'.
    riot (str): String to be converted and stored under 'riot'.
    email (str): String to be converted and stored under 'email'.
    pgp_fingerprint (str): String to be converted and stored under 'pgp_fingerprint'.
    image (str): String to be converted and stored under 'image'.
    info (str): String to be converted and stored under 'info'.
    twitter (str): String to be converted and stored under 'twitter'.

    Returns:
    dict: A dictionary with the specified structure and byte string conversions.

    Raises:
    ValueError: If pgp_fingerprint is not exactly 20 bytes long when encoded.
    """
    if pgp_fingerprint and len(pgp_fingerprint.encode()) != 20:
        raise ValueError("pgp_fingerprint must be exactly 20 bytes long when encoded")

    return {
        "info": {
            "additional": [[]],
            "display": {f"Raw{len(display.encode())}": display.encode()},
            "legal": {f"Raw{len(legal.encode())}": legal.encode()},
            "web": {f"Raw{len(web.encode())}": web.encode()},
            "riot": {f"Raw{len(riot.encode())}": riot.encode()},
            "email": {f"Raw{len(email.encode())}": email.encode()},
            "pgp_fingerprint": pgp_fingerprint.encode() if pgp_fingerprint else None,
            "image": {f"Raw{len(image.encode())}": image.encode()},
            "info": {f"Raw{len(info.encode())}": info.encode()},
            "twitter": {f"Raw{len(twitter.encode())}": twitter.encode()},
        }
    }


def decode_hex_identity_dict(info_dictionary):
    for key, value in info_dictionary.items():
        if isinstance(value, dict):
            item = list(value.values())[0]
            if isinstance(item, str) and item.startswith("0x"):
                try:
                    info_dictionary[key] = bytes.fromhex(item[2:]).decode()
                except UnicodeDecodeError:
                    print(f"Could not decode: {key}: {item}")
            else:
                info_dictionary[key] = item
    return info_dictionary



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import sys
import vana
from rich.prompt import Prompt
from vana.commands.base_command import BaseCommand

from . import defaults

console = vana.__console__


class TransferCommand(BaseCommand):
    """
    Executes the ``transfer`` command to transfer VANA tokens from one account to another on the Vana network.

    This command is used for transactions between different accounts, enabling users to send tokens to other participants on the network.

    Usage:
        The command requires specifying the destination address (public key) and the amount of VANA to be transferred.
        It checks for sufficient balance and prompts for confirmation before proceeding with the transaction.

    Optional arguments:
        - ``--dest`` (str): The destination address for the transfer. This can be in the form of an h160 or secp256k1 public key.
        - ``--amount`` (float): The amount of VANA tokens to transfer.

    The command displays the user's current balance before prompting for the amount to transfer, ensuring transparency and accuracy in the transaction.

    Example usage::

        vanacli wallet transfer --dest 5Dp8... --amount 100

    Note:
        This command is crucial for executing token transfers within the Vana network. Users should verify the destination address and amount before confirming the transaction to avoid errors or loss of funds.
    """

    @staticmethod
    def run(cli: "vana.cli"):
        r"""Transfer token of amount to destination."""
        try:
            chain_manager: "vana.ChainManager" = vana.ChainManager(
                config=cli.config
            )
            TransferCommand._run(cli, chain_manager)
        finally:
            if "chain_manager" in locals():
                chain_manager.close()
                vana.logging.debug("closing chain_manager connection")

    @staticmethod
    def _run(cli: "vana.cli", chain_manager: "vana.ChainManager"):
        r"""Transfer token of amount to destination."""
        wallet = vana.Wallet(config=cli.config)
        chain_manager.transfer(
            wallet=wallet,
            dest=cli.config.dest,
            amount=cli.config.amount,
            wait_for_inclusion=True,
            prompt=not cli.config.no_prompt,
        )

    @staticmethod
    def check_config(config: "vana.Config"):
        if not config.is_set("wallet.name") and not config.no_prompt:
            wallet_name = Prompt.ask("Enter wallet name", default=defaults.wallet.name)
            config.wallet.name = str(wallet_name)

        # Get destination.
        if not config.dest and not config.no_prompt:
            dest = Prompt.ask("Enter destination public key: (h160 or secp256k1)")
            if not vana.utils.is_valid_vana_address_or_public_key(dest):
                sys.exit()
            else:
                config.dest = str(dest)

        # Get current balance and print to user.
        if not config.no_prompt:
            wallet = vana.Wallet(config=config)
            chain_manager = vana.ChainManager(config=config)
            with vana.__console__.status(":satellite: Checking Balance..."):
                account_balance = chain_manager.get_balance(wallet.coldkeypub.to_checksum_address())
                vana.__console__.print(
                    "Balance: [green]{}[/green]".format(account_balance)
                )

        # Get amount.
        if not config.get("amount"):
            amount = Prompt.ask("Enter VANA amount to transfer")
            if not config.no_prompt:
                try:
                    config.amount = float(amount)
                except ValueError:
                    console.print(
                        ":cross_mark:[red] Invalid VANA amount[/red] [bold white]{}[/bold white]".format(
                            amount
                        )
                    )
                    sys.exit()
            else:
                console.print(
                    ":cross_mark:[red] Invalid VANA amount[/red] [bold white]{}[/bold white]".format(
                        amount
                    )
                )
                sys.exit(1)

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        transfer_parser = parser.add_parser(
            "transfer", help="""Transfer VANA between accounts."""
        )
        transfer_parser.add_argument("--dest", dest="dest", type=str, required=False)
        transfer_parser.add_argument(
            "--amount", dest="amount", type=float, required=False
        )

        vana.Wallet.add_args(transfer_parser)
        vana.ChainManager.add_args(transfer_parser)



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import copy
import os
from typing import Optional, Union, Tuple, Dict, overload

from eth_account import Account
from eth_account.signers.local import (
    LocalAccount,
)
from eth_keys.datatypes import PublicKey

from rich.panel import Panel

import vana
from vana.utils.wallet_utils import is_valid_vana_address_or_public_key


def display_private_key_msg(private_key: str, key_type: str):
    """
    Display the private key and a warning message about its sensitivity.

    Args:
        private_key (str): The private key to display.
        key_type (str): Type of the key (coldkey or hotkey).
    """
    vana.__console__.print(Panel.fit(
        f"[bold green]Your {key_type} private key:[/bold green]\n\n"
        f"[yellow]{private_key}[/yellow]\n\n"
        "[bold red]IMPORTANT:[/bold red] Store this private key in a secure (preferably offline) place.\n"
        "Anyone with this private key has full control over the associated account.\n\n"
        f"You can use this private key to import your {key_type} into other wallets.\n"
        "The command to regenerate the key using this private key is:\n"
        f"[cyan]vanacli w regen_{key_type} --seed {private_key}[/cyan]"
    ))

def display_mnemonic_msg(mnemonic: str, key_type: str):
    """
    Display the mnemonic and a warning message to keep the mnemonic safe.

    Args:
        mnemonic (str): Mnemonic string.
        key_type (str): Type of the key (coldkey or hotkey).
    """
    vana.__console__.print(Panel.fit(
        f"[bold green]Your {key_type} mnemonic phrase:[/bold green]\n\n"
        f"[yellow]{mnemonic}[/yellow]\n\n"
        "[bold red]IMPORTANT:[/bold red] Store this mnemonic in a secure (preferably offline) place.\n"
        "Anyone with this mnemonic can regenerate the key and access your tokens.\n\n"
        f"You can use this mnemonic to recreate the {key_type} in case it gets lost.\n"
        "The command to regenerate the key using this mnemonic is:\n"
        f"[cyan]vanacli w regen_{key_type} --mnemonic {mnemonic}[/cyan]"
    ))


class Wallet:
    """
    The wallet class in the handles wallet functionality needed for participating in the Vana network.

    It manages two types of keys: coldkey and hotkey, each serving different purposes in network operations.
    Each wallet contains a coldkey and a hotkey.

    The coldkey is the user's primary key for holding stake in their wallet and is the only way that users
    can access VANA. Coldkeys can hold tokens and should be encrypted on your device.

    The coldkey is the primary key used for securing the wallet's stake in the Vana network and
    is critical for financial transactions like staking and unstaking tokens. It's recommended to keep the
    coldkey encrypted and secure, as it holds the actual tokens.

    The hotkey, in contrast, is used for operational tasks like subscribing to and setting weights in the
    network. It's linked to the coldkey through the metagraph and does not directly hold tokens, thereby
    offering a safer way to interact with the network during regular operations.

    Args:
        name (str): The name of the wallet, used to identify it among possibly multiple wallets.
        path (str): File system path where wallet keys are stored.
        hotkey_str (str): String identifier for the hotkey.
        _hotkey, _coldkey, _coldkeypub (LocalAccount): Internal representations of the hotkey and coldkey.

    Methods:
        create_if_non_existent, create, recreate: Methods to handle the creation of wallet keys.
        get_coldkey, get_hotkey, get_coldkeypub: Methods to retrieve specific keys.
        set_coldkey, set_hotkey, set_coldkeypub: Methods to set or update keys.
        hotkey_file, coldkey_file, coldkeypub_file: Properties that return respective key file objects.
        regenerate_coldkey, regenerate_hotkey, regenerate_coldkeypub: Methods to regenerate keys from different sources.
        config, help, add_args: Utility methods for configuration and assistance.

    Example Usage::

        # Create a new wallet with default coldkey and hotkey names
        my_wallet = wallet()

        # Access hotkey and coldkey
        hotkey = my_wallet.get_hotkey()
        coldkey = my_wallet.get_coldkey()

        # Set a new coldkey
        my_wallet.create_new_coldkey(n_words=24) # number of seed words to use

        # Update wallet hotkey
        my_wallet.set_hotkey(new_hotkey)

        # Print wallet details
        print(my_wallet)

        # Access coldkey property, must use password to unlock
        my_wallet.coldkey
    """

    @classmethod
    def config(cls) -> "vana.Config":
        """
        Get config from the argument parser.

        Returns:
            Config: Config object.
        """
        parser = argparse.ArgumentParser()
        cls.add_args(parser)
        return vana.Config(parser, args=[])

    @classmethod
    def help(cls):
        """
        Print help to stdout.
        """
        parser = argparse.ArgumentParser()
        cls.add_args(parser)
        print(cls.__new__.__doc__)
        parser.print_help()

    @classmethod
    def add_args(cls, parser: argparse.ArgumentParser, prefix: str = None):
        """
        Accept specific arguments from parser.

        Args:
            parser (argparse.ArgumentParser): Argument parser object.
            prefix (str): Argument prefix.
        """
        prefix_str = "" if prefix is None else prefix + "."
        try:
            default_name = os.getenv("WALLET_NAME") or "default"
            default_hotkey = os.getenv("WALLET_NAME") or "default"
            default_path = os.getenv("WALLET_PATH") or "~/.vana/wallets/"
            parser.add_argument(
                "--no_prompt",
                dest="no_prompt",
                action="store_true",
                help="""Set true to avoid prompting the user.""",
                default=False,
            )
            parser.add_argument(
                "--" + prefix_str + "wallet.name",
                required=False,
                default=default_name,
                help="The name of the wallet to unlock for running vana "
                     "(name mock is reserved for mocking this wallet)",
            )
            parser.add_argument(
                "--" + prefix_str + "wallet.hotkey",
                required=False,
                default=default_hotkey,
                help="The name of the wallet's hotkey.",
            )
            parser.add_argument(
                "--" + prefix_str + "wallet.path",
                required=False,
                default=default_path,
                help="The path to your vana wallets",
            )
        except argparse.ArgumentError as e:
            pass

    def __init__(
            self,
            name: str = None,
            hotkey: str = None,
            path: str = None,
            config: "vana.Config" = None,
    ):
        r"""
        Initialize the vana wallet object containing a hot and coldkey.

        Args:
            name (str, optional): The name of the wallet to unlock for running vana. Defaults to ``default``.
            hotkey (str, optional): The name of hotkey used to running the miner. Defaults to ``default``.
            path (str, optional): The path to your vana wallets. Defaults to ``~/.vana/wallets/``.
            config (Config, optional): Wallet.config(). Defaults to ``None``.
        """
        # Fill config from passed args using command line defaults.
        if config is None:
            config = Wallet.config()
        self.config = copy.deepcopy(config)
        self.config.wallet.name = name or self.config.wallet.get(
            "name", vana.defaults.wallet.name
        )
        self.config.wallet.hotkey = hotkey or self.config.wallet.get(
            "hotkey", vana.defaults.wallet.hotkey
        )
        self.config.wallet.path = path or self.config.wallet.get(
            "path", vana.defaults.wallet.path
        )

        self.name = self.config.wallet.name
        self.path = self.config.wallet.path
        self.hotkey_str = self.config.wallet.hotkey

        self._hotkey = None
        self._coldkey = None
        self._coldkeypub = None
        self._mnemonics = {}

        # Create necessary directories
        self._create_wallet_directories()

        # Check for and handle environment variable based wallet initialization
        self._init_from_environment()

    def _create_wallet_directories(self):
        """Create the necessary wallet directory structure if it doesn't exist."""
        # Create main wallet directory
        wallet_path = os.path.expanduser(os.path.join(self.path, self.name))
        # Create hotkeys directory
        hotkeys_path = os.path.join(wallet_path, "hotkeys")

        # Create directories if they don't exist
        os.makedirs(hotkeys_path, exist_ok=True)

    def _init_from_environment(self):
        """
        Initialize wallet components from environment variables if they exist.
        Currently supports hotkey initialization from HOTKEY_MNEMONIC.
        Creates the wallet files on disk without password protection.
        """
        hotkey_mnemonic = os.getenv("HOTKEY_MNEMONIC")
        if hotkey_mnemonic:
            try:
                vana.logging.info("Found HOTKEY_MNEMONIC environment variable. Initializing hotkey...")

                # Enable mnemonic features
                Account.enable_unaudited_hdwallet_features()

                # Create account from mnemonic
                account = Account.from_mnemonic(hotkey_mnemonic)

                # Create the hotkey file
                wallet_path = os.path.expanduser(os.path.join(self.path, self.name))
                hotkey_path = os.path.join(wallet_path, "hotkeys", self.hotkey_str)

                # Create the keyfile
                keyfile = vana.keyfile(path=hotkey_path)

                # Set the hotkey without password protection
                keyfile.set_keypair(
                    keypair=account,
                    encrypt=False,  # No encryption for environment-based keys
                    overwrite=True  # Always overwrite when using environment variables
                )

                # Update the wallet's hotkey reference
                self._hotkey = account

                vana.logging.success(
                    f"Successfully initialized hotkey from environment variable and created file at {hotkey_path}"
                )

            except Exception as e:
                vana.logging.error(f"Failed to initialize hotkey from environment variable: {str(e)}")
                # Don't raise the exception - allow fallback to normal initialization
                pass

    def __str__(self):
        """
        Returns the string representation of the Wallet object.

        Returns:
            str: The string representation.
        """
        return "wallet({}, {}, {})".format(self.name, self.hotkey_str, self.path)

    def __repr__(self):
        """
        Returns the string representation of the wallet object.

        Returns:
            str: The string representation.
        """
        return self.__str__()

    def create_if_non_existent(
            self, coldkey_use_password: bool = True, hotkey_use_password: bool = False
    ) -> "Wallet":
        """
        Checks for existing coldkeypub and hotkeys, and creates them if non-existent.

        Args:
            coldkey_use_password (bool, optional): Whether to use a password for coldkey. Defaults to ``True``.
            hotkey_use_password (bool, optional): Whether to use a password for hotkey. Defaults to ``False``.

        Returns:
            wallet: The wallet object.
        """
        return self.create(coldkey_use_password, hotkey_use_password)

    def create(
            self, coldkey_use_password: bool = True, hotkey_use_password: bool = False
    ) -> "Wallet":
        """
        Checks for existing coldkeypub and hotkeys, and creates them if non-existent.

        Args:
            coldkey_use_password (bool, optional): Whether to use a password for coldkey. Defaults to ``True``.
            hotkey_use_password (bool, optional): Whether to use a password for hotkey. Defaults to ``False``.

        Returns:
            wallet: The wallet object.
        """
        # ---- Setup Wallet. ----
        if (
                not self.coldkey_file.exists_on_device()
                and not self.coldkeypub_file.exists_on_device()
        ):
            self.create_new_coldkey(n_words=12, use_password=coldkey_use_password)
        if not self.hotkey_file.exists_on_device():
            self.create_new_hotkey(n_words=12, use_password=hotkey_use_password)
        return self

    def recreate(
            self, coldkey_use_password: bool = True, hotkey_use_password: bool = False
    ) -> "Wallet":
        """
        Checks for existing coldkeypub and hotkeys and creates them if non-existent.

        Args:
            coldkey_use_password (bool, optional): Whether to use a password for coldkey. Defaults to ``True``.
            hotkey_use_password (bool, optional): Whether to use a password for hotkey. Defaults to ``False``.

        Returns:
            wallet: The wallet object.
        """
        self.create_new_coldkey(n_words=12, use_password=coldkey_use_password)
        self.create_new_hotkey(n_words=12, use_password=hotkey_use_password)
        return self

    @property
    def hotkey_file(self) -> "vana.keyfile":
        """
        Property that returns the hotkey file.

        Returns:
            keyfile: The hotkey file.
        """
        wallet_path = os.path.expanduser(os.path.join(self.path, self.name))
        hotkey_path = os.path.join(wallet_path, "hotkeys", self.hotkey_str)
        return vana.keyfile(path=hotkey_path)

    @property
    def coldkey_file(self) -> "vana.keyfile":
        """
        Property that returns the coldkey file.

        Returns:
            keyfile: The coldkey file.
        """
        wallet_path = os.path.expanduser(os.path.join(self.path, self.name))
        coldkey_path = os.path.join(wallet_path, "coldkey")
        return vana.keyfile(path=coldkey_path)

    @property
    def coldkeypub_file(self) -> "vana.keyfile":
        """
        Property that returns the coldkeypub file.

        Returns:
            keyfile: The coldkeypub file.
        """
        wallet_path = os.path.expanduser(os.path.join(self.path, self.name))
        coldkeypub_path = os.path.join(wallet_path, "coldkeypub.txt")
        return vana.keyfile(path=coldkeypub_path)

    def set_hotkey(
            self,
            account: LocalAccount,
            encrypt: bool = False,
            overwrite: bool = False,
    ) -> "vana.keyfile":
        """
        Sets the hotkey for the wallet.

        Args:
            account (LocalAccount): The hotkey account.
            encrypt (bool, optional): Whether to encrypt the hotkey. Defaults to ``False``.
            overwrite (bool, optional): Whether to overwrite an existing hotkey. Defaults to ``False``.

        Returns:
            keyfile: The hotkey file.
        """
        self._hotkey = account
        self.hotkey_file.set_keypair(account, encrypt=encrypt, overwrite=overwrite)
        return self.hotkey_file

    def set_coldkeypub(
            self,
            account: LocalAccount,
            encrypt: bool = False,
            overwrite: bool = False,
    ) -> "vana.keyfile":
        """
        Sets the coldkeypub for the wallet.

        Args:
            account (LocalAccount): The coldkeypub account.
            encrypt (bool, optional): Whether to encrypt the coldkeypub. Defaults to ``False``.
            overwrite (bool, optional): Whether to overwrite an existing coldkeypub. Defaults to ``False``.

        Returns:
            keyfile: The coldkeypub file.
        """
        self._coldkeypub = account._key_obj.public_key
        self.coldkeypub_file.set_keypair(
            self._coldkeypub, encrypt=encrypt, overwrite=overwrite
        )
        return self.coldkeypub_file

    def set_coldkey(
            self,
            account: LocalAccount,
            encrypt: bool = True,
            overwrite: bool = False,
    ) -> "vana.keyfile":
        """
        Sets the coldkey for the wallet.

        Args:
            account (LocalAccount): The coldkey account.
            encrypt (bool, optional): Whether to encrypt the coldkey. Defaults to ``True``.
            overwrite (bool, optional): Whether to overwrite an existing coldkey. Defaults to ``False``.

        Returns:
            keyfile: The coldkey file.
        """
        self._coldkey = account
        self.coldkey_file.set_keypair(
            self._coldkey, encrypt=encrypt, overwrite=overwrite
        )
        return self.coldkey_file

    def get_coldkey(self, password: str = None) -> LocalAccount:
        """
        Gets the coldkey from the wallet.

        Args:
            password (str, optional): The password to decrypt the coldkey. Defaults to ``None``.

        Returns:
            LocalAccount: The coldkey account.
        """
        return self.coldkey_file.get_keypair(password=password)

    def get_hotkey(self, password: str = None) -> LocalAccount:
        """
        Gets the hotkey from the wallet.

        Args:
            password (str, optional): The password to decrypt the hotkey. Defaults to ``None``.

        Returns:
            LocalAccount: The hotkey account.
        """
        return self.hotkey_file.get_keypair(password=password)

    def get_hotkey_public_key(self) -> str:
        """
        Gets the public key of the hotkey as a hexadecimal string.

        Returns:
            str: The public key of the hotkey in hexadecimal format.

        Raises:
            AttributeError: If the hotkey is not set.
        """
        if self._hotkey is None:
            self._hotkey = self.hotkey_file.keypair

        if self._hotkey is None:
            raise AttributeError("Hotkey is not set. Please ensure the hotkey is properly initialized.")

        # Get the public key as bytes
        public_key_bytes = self._hotkey._key_obj.public_key.to_bytes()

        # Convert the bytes to a hexadecimal string
        return '0x' + public_key_bytes.hex()

    def get_coldkeypub(self, password: str = None) -> PublicKey:
        """
        Gets the coldkeypub from the wallet.

        Args:
            password (str, optional): The password to decrypt the coldkeypub. Defaults to ``None``.

        Returns:
            LocalAccount: The coldkeypub account.
        """
        return self.coldkeypub_file.get_keypair(password=password)

    @property
    def hotkey(self) -> LocalAccount:
        r"""Loads the hotkey from wallet.path/wallet.name/hotkeys/wallet.hotkey or raises an error.

        Returns:
            hotkey (LocalAccount):
                hotkey loaded from config arguments.
        Raises:
            KeyFileError: Raised if the file is corrupt of non-existent.
            CryptoKeyError: Raised if the user enters an incorrect password for an encrypted keyfile.
        """
        if self._hotkey is None:
            self._hotkey = self.hotkey_file.keypair
        return self._hotkey

    @property
    def coldkey(self) -> LocalAccount:
        r"""Loads the hotkey from wallet.path/wallet.name/coldkey or raises an error.

        Returns:
            coldkey (LocalAccount): coldkey loaded from config arguments.
        Raises:
            KeyFileError: Raised if the file is corrupt of non-existent.
            CryptoKeyError: Raised if the user enters an incorrect password for an encrypted keyfile.
        """
        if self._coldkey is None:
            self._coldkey = self.coldkey_file.keypair
        return self._coldkey

    @property
    def coldkeypub(self) -> PublicKey:
        r"""Loads the coldkeypub from wallet.path/wallet.name/coldkeypub.txt or raises an error.

        Returns:
            coldkeypub (LocalAccount): coldkeypub loaded from config arguments.
        Raises:
            KeyFileError: Raised if the file is corrupt of non-existent.
            CryptoKeyError: Raised if the user enters an incorrect password for an encrypted keyfile.
        """
        if self._coldkeypub is None:
            self._coldkeypub = self.coldkeypub_file.keypair
        return self._coldkeypub

    @property
    def coldkeypub_str(self) -> str | None:
        """
        Returns the coldkeypub address as a string based on its type
        """
        if str(type(self.coldkeypub)) == "<class 'eth_account.signers.local.LocalAccount'>":
            return self.coldkeypub.address
        if str(type(self.coldkeypub)) == "<class 'eth_keys.datatypes.PublicKey'>":
            return str(self.coldkeypub)
        return None

    def create_new_coldkey(
            self,
            n_words: int = 12,
            use_password: bool = True,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        """Creates a new coldkey, optionally encrypts it with the user-provided password and saves to disk.

        Args:
            n_words: (int, optional):
                Number of mnemonic words to use.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Determines if this operation overwrites the coldkey under the same path ``<wallet path>/<wallet name>/coldkey``.
        Returns:
            wallet (Wallet):
                This object with newly created coldkey.
        """
        Account.enable_unaudited_hdwallet_features()
        account, mnemonic = Account.create_with_mnemonic()
        if not suppress:
            display_mnemonic_msg(mnemonic, "coldkey")
        self._mnemonics["coldkey"] = mnemonic
        self.set_coldkey(account, encrypt=use_password, overwrite=overwrite)
        self.set_coldkeypub(account, overwrite=overwrite)
        return self

    def new_hotkey(
            self,
            n_words: int = 12,
            use_password: bool = False,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        """Creates a new hotkey, optionally encrypts it with the user-provided password and saves to disk.

        Args:
            n_words: (int, optional):
                Number of mnemonic words to use.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Determines if this operation overwrites the hotkey under the same path ``<wallet path>/<wallet name>/hotkeys/<hotkey>``.
        Returns:
            wallet (Wallet):
                This object with newly created hotkey.
        """
        return self.create_new_hotkey(n_words, use_password, overwrite, suppress)

    def create_new_hotkey(
            self,
            n_words: int = 12,
            use_password: bool = False,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        """Creates a new hotkey, optionally encrypts it with the user-provided password and saves to disk.

        Args:
            n_words: (int, optional):
                Number of mnemonic words to use.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the hotkey under the same path <wallet path>/<wallet name>/hotkeys/<hotkey>
        Returns:
            wallet (Wallet):
                This object with newly created hotkey.
        """
        Account.enable_unaudited_hdwallet_features()
        account, mnemonic = Account.create_with_mnemonic()
        if not suppress:
            display_mnemonic_msg(mnemonic, "hotkey")
        self._mnemonics["hotkey"] = mnemonic
        self.set_hotkey(account, encrypt=use_password, overwrite=overwrite)
        return self

    def regenerate_coldkeypub(
            self,
            h160_address: Optional[str] = None,
            public_key: Optional[Union[str, bytes]] = None,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        """Regenerates the coldkeypub from the passed ``h160_address`` or public_key and saves the file. Requires either ``h160_address`` or public_key to be passed.

        Args:
            h160_address: (str, optional):
                Address as ``h160`` string.
            public_key: (str | bytes, optional):
                Public key as hex string or bytes.
            overwrite (bool, optional) (default: False):
                Determines if this operation overwrites the coldkeypub (if exists) under the same path ``<wallet path>/<wallet name>/coldkeypub``.
        Returns:
            wallet (Wallet):
                Newly re-generated wallet with coldkeypub.

        """
        if h160_address is None and public_key is None:
            raise ValueError("Either h160_address or public_key must be passed")

        if not is_valid_vana_address_or_public_key(
                h160_address if h160_address is not None else public_key
        ):
            raise ValueError(
                f"Invalid {'h160_address' if h160_address is not None else 'public_key'}"
            )

        if h160_address is not None:
            account = Account.from_key(h160_address)
        else:
            account = Account.from_key(public_key)

        # No need to encrypt the public key
        self.set_coldkeypub(account, overwrite=overwrite)

        return self

    # Short name for regenerate_coldkeypub
    regen_coldkeypub = regenerate_coldkeypub

    @overload
    def regenerate_coldkey(
            self,
            mnemonic: Optional[Union[list, str]] = None,
            use_password: bool = True,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        ...

    @overload
    def regenerate_coldkey(
            self,
            seed: Optional[str] = None,
            use_password: bool = True,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        ...

    @overload
    def regenerate_coldkey(
            self,
            json: Optional[Tuple[Union[str, Dict], str]] = None,
            use_password: bool = True,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        ...

    def regenerate_coldkey(
            self,
            use_password: bool = True,
            overwrite: bool = False,
            suppress: bool = False,
            **kwargs,
    ) -> "Wallet":
        """Regenerates the coldkey from the passed mnemonic or seed, or JSON encrypts it with the user's password and saves the file.

        Args:
            mnemonic: (Union[list, str], optional):
                Key mnemonic as list of words or string space separated words.
            seed: (str, optional):
                Seed as hex string.
            json: (Tuple[Union[str, Dict], str], optional):
                Restore from encrypted JSON backup as ``(json_data: Union[str, Dict], passphrase: str)``.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Determines if this operation overwrites the coldkey under the same path ``<wallet path>/<wallet name>/coldkey``.
        Returns:
            wallet (Wallet):
                This object with newly created coldkey.

        Note:
            Uses priority order: ``mnemonic > seed > json``.

        """
        if len(kwargs) == 0:
            raise ValueError("Must pass either mnemonic, seed, or json")

        # Get from kwargs
        mnemonic = kwargs.get("mnemonic", None)
        seed = kwargs.get("seed", None)
        json = kwargs.get("json", None)

        Account.enable_unaudited_hdwallet_features()

        if mnemonic is None and seed is None and json is None:
            raise ValueError("Must pass either mnemonic, seed, or json")
        if mnemonic is not None:
            if isinstance(mnemonic, str):
                mnemonic = mnemonic.split()
            # TODO: only support EVM-compatible lengths
            if len(mnemonic) not in [12, 15, 18, 21, 24]:
                raise ValueError(
                    "Mnemonic has invalid size. This should be 12,15,18,21 or 24 words"
                )
            account = Account.from_mnemonic(" ".join(mnemonic))
            if not suppress:
                display_mnemonic_msg(" ".join(mnemonic), "coldkey")
        elif seed is not None:
            account = Account.from_key(seed)
        else:
            # json is not None
            if (
                    not isinstance(json, tuple)
                    or len(json) != 2
                    or not isinstance(json[0], (str, dict))
                    or not isinstance(json[1], str)
            ):
                raise ValueError(
                    "json must be a tuple of (json_data: str | Dict, passphrase: str)"
                )

            json_data, passphrase = json
            account = Account.decrypt(json_data, passphrase)

        self.set_coldkey(account, encrypt=use_password, overwrite=overwrite)
        self.set_coldkeypub(account, overwrite=overwrite)
        return self

    # Short name for regenerate_coldkey
    regen_coldkey = regenerate_coldkey

    @overload
    def regenerate_hotkey(
            self,
            mnemonic: Optional[Union[list, str]] = None,
            use_password: bool = False,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        ...

    @overload
    def regenerate_hotkey(
            self,
            seed: Optional[str] = None,
            use_password: bool = False,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        ...

    @overload
    def regenerate_hotkey(
            self,
            json: Optional[Tuple[Union[str, Dict], str]] = None,
            use_password: bool = False,
            overwrite: bool = False,
            suppress: bool = False,
    ) -> "Wallet":
        ...

    def regenerate_hotkey(
            self,
            use_password: bool = False,
            overwrite: bool = False,
            suppress: bool = False,
            **kwargs,
    ) -> "Wallet":
        """Regenerates the hotkey from the passed mnemonic or seed, or JSON encrypts it with the user's password and saves the file.

        Args:
            mnemonic: (Union[list, str], optional):
                Key mnemonic as list of words or string space separated words.
            seed: (str, optional):
                Seed as hex string.
            json: (Tuple[Union[str, Dict], str], optional):
                Restore from encrypted JSON backup as ``(json_data: Union[str, Dict], passphrase: str)``.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Determines if this operation overwrites the hotkey under the same path ``<wallet path>/<wallet name>/hotkeys/<hotkey>``.
        Returns:
            wallet (Wallet):
                This object with newly created hotkey.

        """
        if len(kwargs) == 0:
            raise ValueError("Must pass either mnemonic, seed, or json")

        # Get from kwargs
        mnemonic = kwargs.get("mnemonic", None)
        seed = kwargs.get("seed", None)
        json = kwargs.get("json", None)

        Account.enable_unaudited_hdwallet_features()

        if mnemonic is None and seed is None and json is None:
            raise ValueError("Must pass either mnemonic, seed, or json")

        if mnemonic is not None:
            if isinstance(mnemonic, str):
                mnemonic = mnemonic.split()
            # TODO: only support EVM-compatible lengths
            if len(mnemonic) not in [12, 15, 18, 21, 24]:
                raise ValueError(
                    "Mnemonic has invalid size. This should be 12,15,18,21 or 24 words"
                )
            account = Account.from_mnemonic(" ".join(mnemonic))
            if not suppress:
                display_mnemonic_msg(" ".join(mnemonic), "hotkey")
        elif seed is not None:
            account = Account.from_key(seed)
        else:
            # json is not None
            if (
                    not isinstance(json, tuple)
                    or len(json) != 2
                    or not isinstance(json[0], (str, dict))
                    or not isinstance(json[1], str)
            ):
                raise ValueError(
                    "json must be a tuple of (json_data: str | Dict, passphrase: str)"
                )

            json_data, passphrase = json
            account = Account.decrypt(json_data, passphrase)

        self.set_hotkey(account, encrypt=use_password, overwrite=overwrite)
        return self

    # Short name for regenerate_hotkey
    regen_hotkey = regenerate_hotkey



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

__version__ = "0.45.0"

import rich

from .config import Config

version_split = __version__.split(".")
__version_as_int__: int = (
        (100 * int(version_split[0]))
        + (10 * int(version_split[1]))
        + (1 * int(version_split[2]))
)

# Rich console.
__console__ = rich.get_console()
__use_console__ = True


def turn_console_off():
    global __use_console__
    global __console__
    from io import StringIO

    __use_console__ = False
    rich.reconfigure(file=StringIO(), stderr=False)


def turn_console_on():
    global __use_console__
    global __console__
    __use_console__ = True
    __console__ = rich.get_console()


# TODO: turn_console_on() isn't working so this is commented out for now.
# turn_console_off()

from .errors import (
    BlacklistedException,
    ChainConnectionError,
    ChainError,
    ChainQueryError,
    ChainTransactionError,
    IdentityError,
    InternalServerError,
    InvalidRequestNameError,
    KeyFileError,
    MetadataError,
    NominationError,
    NotDelegateError,
    NotRegisteredError,
    NotVerifiedException,
    PostProcessException,
    PriorityException,
    RegistrationError,
    RunException,
    StakeError,
    TransferError,
    UnstakeError,
)
from .keyfile import (
    serialized_keypair_to_keyfile_data,
    deserialize_keypair_from_keyfile_data,
    validate_password,
    ask_password_to_encrypt,
    keyfile_data_is_encrypted_nacl,
    keyfile_data_is_encrypted_ansible,
    keyfile_data_is_encrypted_legacy,
    keyfile_data_is_encrypted,
    keyfile_data_encryption_method,
    legacy_encrypt_keyfile_data,
    encrypt_keyfile_data,
    get_coldkey_password_from_environment,
    decrypt_keyfile_data,
    keyfile,
    Mockkeyfile,
)
from .wallet import Wallet
from .utils import (
    hash,
    wallet_utils,
)
from .chain_data import (
    NodeServerInfo
)
from .chain_manager import ChainManager
from .cli import cli as cli, COMMANDS as ALL_COMMANDS
from .logging import logging
from .state import State
from .message import Message, TerminalInfo
from .node_server import NodeServer
from .node_client import NodeClient
from .client import Client


# Logging helpers.
def trace(on: bool = True):
    logging.set_trace(on)


def debug(on: bool = True):
    logging.set_debug(on)


__networks__ = ["vana", "islander", "maya", "satori", "moksha", "local", "test", "archive"]

__vana_entrypoint__ = "https://rpc.vana.org"

__islander_entrypoint__ = "https://rpc.islander.vana.org"

__maya_entrypoint__ = "https://rpc.maya.vana.org"

__satori_entrypoint__ = "https://rpc.satori.vana.org"

__moksha_entrypoint__ = "https://rpc.moksha.vana.org"

__archive_entrypoint__ = "wss://archive.vana.org:443"

__local_entrypoint__ = "ws://127.0.0.1:9944"

block_explorer_tx_templates = {
    "vana": "https://vanascan.io/tx/{}",
    "moksha": "https://moksha.vanascan.io/tx/{}",
    "satori": "https://satori.vanascan.io/tx/{}",
    "islander": "https://islander.vanascan.io/tx/{}",
    "maya": "https://maya.vanascan.io/tx/{}",
}

configs = [
    NodeServer.config(),
    Wallet.config(),
    ChainManager.config(),
    logging.get_config(),
]
defaults = Config.merge_all(configs)

banner = r"""_/\\\________/\\\_____/\\\\\\\\\_____/\\\\\_____/\\\_____/\\\\\\\\\\\___
_\/\\\_______\/\\\___/\\\\\\\\\\\\\__\/\\\\\\___\/\\\___/\\\\\\\\\\\\\__
_\//\\\______/\\\___/\\\/////////\\\_\/\\\/\\\__\/\\\__/\\\/////////\\\_
__\//\\\____/\\\___\/\\\_______\/\\\_\/\\\//\\\_\/\\\_\/\\\_______\/\\\_
___\//\\\__/\\\____\/\\\\\\\\\\\\\\\_\/\\\\//\\\\/\\\_\/\\\\\\\\\\\\\\\_
____\//\\\/\\\_____\/\\\/////////\\\_\/\\\_\//\\\/\\\_\/\\\/////////\\\_
_____\//\\\\\______\/\\\_______\/\\\_\/\\\__\//\\\\\\_\/\\\_______\/\\\_
______\//\\\_______\/\\\_______\/\\\_\/\\\___\//\\\\\_\/\\\_______\/\\\_
_______\///________\///________\///__\///_____\/////__\///________\///__
"""

import sys
import os


def is_cli_context():
    """
    Determine if the current context is a CLI invocation by checking how the script is being executed.
    """
    if sys.argv[0] == '-c':
        return True

    main_module = sys.modules.get('__main__')
    if main_module:
        module_name = getattr(main_module, '__name__', '')
        file_name = os.path.basename(getattr(main_module, '__file__', ''))
        return 'cli' in module_name.lower() or 'cli' in file_name.lower()

    return False


if not is_cli_context():
    print(banner)



# The MIT License (MIT)
# Copyright © 2024 Corsali, Inc. dba Vana

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import atexit
import copy
import logging as stdlogging
import multiprocessing as mp
import os
import sys
from logging.handlers import QueueHandler, QueueListener, RotatingFileHandler
from typing import NamedTuple

from statemachine import StateMachine, State

import vana
from vana.logging.defines import (
    TRACE_LOG_FORMAT,
    DATE_FORMAT,
    VANA_LOGGER_NAME,
    DEFAULT_LOG_FILE_NAME,
    DEFAULT_MAX_ROTATING_LOG_FILE_SIZE,
    DEFAULT_LOG_BACKUP_COUNT,
)
from vana.logging.format import StreamFormatter, FileFormatter
from vana.logging.helpers import all_loggers


class LoggingConfig(NamedTuple):
    debug: bool
    trace: bool
    record_log: bool
    logging_dir: str


class LoggingMachine(StateMachine):
    """
    Handles logger states for vana and 3rd party libraries
    """

    Default = State(initial=True)
    Debug = State()
    Trace = State()
    Disabled = State()

    enable_default = (
            Debug.to(Default)
            | Trace.to(Default)
            | Disabled.to(Default)
            | Default.to(Default)
    )

    enable_trace = (
            Default.to(Trace) | Debug.to(Trace) | Disabled.to(Trace) | Trace.to(Trace)
    )

    enable_debug = (
            Default.to(Debug) | Trace.to(Debug) | Disabled.to(Debug) | Debug.to(Debug)
    )

    disable_trace = Trace.to(Default)

    disable_debug = Debug.to(Default)

    disable_logging = (
            Trace.to(Disabled)
            | Debug.to(Disabled)
            | Default.to(Disabled)
            | Disabled.to(Disabled)
    )

    def __init__(self, config: "vana.Config", name: str = VANA_LOGGER_NAME):
        # basics
        super(LoggingMachine, self).__init__()
        self._queue = mp.Queue(-1)
        self._name = name
        self._config = config

        # Formatters
        #
        # In the future, this may be expanded to a dictionary mapping handler
        # types to their respective formatters.
        self._stream_formatter = StreamFormatter()
        self._file_formatter = FileFormatter(TRACE_LOG_FORMAT, DATE_FORMAT)

        # start with handlers for the QueueListener.
        #
        # In the future, we may want to add options to introduce other handlers
        # for things like log aggregation by external services.
        self._handlers = self._configure_handlers(config)

        # configure and start the queue listener
        self._listener = self._create_and_start_listener(self._handlers)

        # set up all the loggers
        self._logger = self._initialize_logger(name, config)
        self.disable_third_party_loggers()

    def _configure_handlers(self, config) -> list[stdlogging.Handler]:
        handlers = list()

        # stream handler, a given
        stream_handler = stdlogging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(self._stream_formatter)
        handlers.append(stream_handler)

        # file handler, maybe
        if config.record_log and config.logging_dir:
            logfile = os.path.abspath(
                os.path.join(config.logging_dir, DEFAULT_LOG_FILE_NAME)
            )
            file_handler = self._create_file_handler(logfile)
            handlers.append(file_handler)
        return handlers

    def get_config(self):
        return self._config

    def set_config(self, config):
        """
        Set config after initialization, if desired.
        """
        self._config = config
        if config.logging_dir and config.record_log:
            logfile = os.path.abspath(
                os.path.join(config.logging_dir, DEFAULT_LOG_FILE_NAME)
            )
            self._enable_file_logging(logfile)
        if config.trace:
            self.enable_trace()
        elif config.debug:
            self.enable_debug()

    def _create_and_start_listener(self, handlers):
        """
        A listener to receive and publish log records.

        This listener receives records from a queue populated by the main opendata
        logger, as well as 3rd party loggers.
        """

        listener = QueueListener(self._queue, *handlers, respect_handler_level=True)
        listener.start()
        atexit.register(listener.stop)
        return listener

    def get_queue(self):
        """
        Get the queue the QueueListener is publishing from.

        To set up logging in a separate process, a QueueHandler must
        be added to all the desired loggers.
        """
        return self._queue

    def _initialize_logger(self, name, config):
        """
        Initialize logging for vana.

        Since the initial state is Default, logging level for the module logger
        is INFO, and all third-party loggers are silenced. Subsequent state
        transitions will handle all logger outputs.
        """
        logger = stdlogging.getLogger(name)
        queue_handler = QueueHandler(self._queue)
        logger.addHandler(queue_handler)
        return logger

    def _create_file_handler(self, logfile: str):
        file_handler = RotatingFileHandler(
            logfile,
            maxBytes=DEFAULT_MAX_ROTATING_LOG_FILE_SIZE,
            backupCount=DEFAULT_LOG_BACKUP_COUNT,
        )
        file_handler.setFormatter(self._file_formatter)
        file_handler.setLevel(stdlogging.TRACE)
        return file_handler

    def enable_third_party_loggers(self):
        for logger in all_loggers():
            if logger.name == self._name:
                continue
            queue_handler = QueueHandler(self._queue)
            logger.addHandler(queue_handler)
            logger.setLevel(self._logger.level)

    def disable_third_party_loggers(self):
        # remove all handlers
        for logger in all_loggers():
            if logger.name == self._name:
                continue
            for handler in logger.handlers:
                logger.removeHandler(handler)

    def _enable_file_logging(self, logfile: str):
        # preserve idempotency; do not create extra filehandlers
        # if one already exists
        if any(
                [isinstance(handler, RotatingFileHandler) for handler in self._handlers]
        ):
            return
        file_handler = self._create_file_handler(logfile)
        self._handlers.append(file_handler)
        self._listener.handlers = tuple(self._handlers)

    # state transitions
    def before_transition(self, event, state):
        self._listener.stop()

    def after_transition(self, event, state):
        self._listener.start()

    # Default Logging
    def before_enable_default(self):
        self._logger.info(f"Enabling default logging.")
        self._logger.setLevel(stdlogging.INFO)
        for logger in all_loggers():
            if logger.name == self._name:
                continue
            logger.setLevel(stdlogging.CRITICAL)

    def after_enable_default(self):
        pass

    # Trace
    def before_enable_trace(self):
        self._logger.info("Enabling trace.")
        self._stream_formatter.set_trace(True)
        for logger in all_loggers():
            logger.setLevel(stdlogging.TRACE)

    def after_enable_trace(self):
        self._logger.info("Trace enabled.")

    def before_disable_trace(self):
        self._logger.info(f"Disabling trace.")
        self._stream_formatter.set_trace(False)
        self.enable_default()

    def after_disable_trace(self):
        self._logger.info("Trace disabled.")

    # Debug
    def before_enable_debug(self):
        self._logger.info("Enabling debug.")
        self._stream_formatter.set_trace(True)
        for logger in all_loggers():
            logger.setLevel(stdlogging.DEBUG)

    def after_enable_debug(self):
        self._logger.info("Debug enabled.")

    def before_disable_debug(self):
        self._logger.info("Disabling debug.")
        self._stream_formatter.set_trace(False)
        self.enable_default()

    def after_disable_debug(self):
        self._logger.info("Debug disabled.")

    # Disable Logging
    def before_disable_logging(self):
        self._logger.info("Disabling logging.")
        self._stream_formatter.set_trace(False)

        for logger in all_loggers():
            logger.setLevel(stdlogging.CRITICAL)

    # Required API
    # support log commands for API backwards compatibility
    @property
    def __trace_on__(self):
        return self.current_state_value == "Trace"

    def trace(self, msg="", prefix="", sufix="", *args, **kwargs):
        msg = f"{prefix} - {msg} - {sufix}"
        self._logger.trace(msg, *args, **kwargs)

    def debug(self, msg="", prefix="", sufix="", *args, **kwargs):
        msg = f"{prefix} - {msg} - {sufix}"
        self._logger.debug(msg, *args, **kwargs)

    def info(self, msg="", prefix="", sufix="", *args, **kwargs):
        msg = f"{prefix} - {msg} - {sufix}"
        self._logger.info(msg, *args, **kwargs)

    def success(self, msg="", prefix="", sufix="", *args, **kwargs):
        msg = f"{prefix} - {msg} - {sufix}"
        self._logger.success(msg, *args, **kwargs)

    def warning(self, msg="", prefix="", sufix="", *args, **kwargs):
        msg = f"{prefix} - {msg} - {sufix}"
        self._logger.warning(msg, *args, **kwargs)

    def error(self, msg="", prefix="", sufix="", *args, **kwargs):
        msg = f"{prefix} - {msg} - {sufix}"
        self._logger.error(msg, *args, **kwargs)

    def critical(self, msg="", prefix="", sufix="", *args, **kwargs):
        msg = f"{prefix} - {msg} - {sufix}"
        self._logger.critical(msg, *args, **kwargs)

    def exception(self, msg="", prefix="", sufix="", *args, **kwargs):
        msg = f"{prefix} - {msg} - {sufix}"
        self._logger.exception(msg, *args, **kwargs)

    def on(self):
        self._logger.info("Logging enabled.")
        self.enable_default()

    def off(self):
        self.disable_logging()

    def set_debug(self, on: bool = True):
        if on and not self.current_state_value == "Debug":
            self.enable_debug()
        elif not on:
            if self.current_state_value == "Debug":
                self.disable_debug()

    def set_trace(self, on: bool = True):
        if on and not self.current_state_value == "Trace":
            self.enable_trace()
        elif not on:
            if self.current_state_value == "Trace":
                self.disable_trace()

    def get_level(self):
        return self._logger.level

    def check_config(self, config: "vana.Config"):
        assert config.logging

    def help(self):
        pass

    @classmethod
    def add_args(cls, parser: argparse.ArgumentParser, prefix: str = None):
        """Accept specific arguments fro parser"""
        prefix_str = "" if prefix == None else prefix + "."
        try:
            default_logging_debug = os.getenv("LOGGING_DEBUG") or False
            default_logging_trace = os.getenv("LOGGING_TRACE") or False
            default_logging_record_log = os.getenv("LOGGING_RECORD_LOG") or False
            default_logging_logging_dir = (
                    os.getenv("LOGGING_LOGGING_DIR") or "~/.vana/miners"
            )
            parser.add_argument(
                "--" + prefix_str + "logging.debug",
                action="store_true",
                help="""Turn on vana debugging information""",
                default=default_logging_debug,
            )
            parser.add_argument(
                "--" + prefix_str + "logging.trace",
                action="store_true",
                help="""Turn on vana trace level information""",
                default=default_logging_trace,
            )
            parser.add_argument(
                "--" + prefix_str + "logging.record_log",
                action="store_true",
                help="""Turns on logging to file.""",
                default=default_logging_record_log,
            )
            parser.add_argument(
                "--" + prefix_str + "logging.logging_dir",
                type=str,
                help="Logging default root directory.",
                default=default_logging_logging_dir,
            )
        except argparse.ArgumentError:
            # re-parsing arguments.
            pass

    @classmethod
    def config(cls):
        """Get config from the argument parser.

        Return:
            config object
        """
        parser = argparse.ArgumentParser()
        cls.add_args(parser)
        return vana.Config(parser, args=[])

    def __call__(
            self,
            config: "vana.Config" = None,
            debug: bool = None,
            trace: bool = None,
            record_log: bool = None,
            logging_dir: str = None,
    ):
        if config is not None:
            cfg = copy.deepcopy(config)
            if debug is not None:
                cfg.debug = debug
            elif trace is not None:
                cfg.trace = trace
            if record_log is not None:
                cfg.record_log = record_log
            if logging_dir is not None:
                cfg.logging_dir = logging_dir
        else:
            cfg = LoggingConfig(
                debug=debug, trace=trace, record_log=record_log, logging_dir=logging_dir
            )
        self.set_config(cfg)
